import os
import re
import uuid
from abc import abstractmethod
from logging import warn
from uuid import uuid4
from typing import *

from process_bigraph.composite import Emitter, Process
from pymongo import ASCENDING, MongoClient
from pymongo.database import Database


HISTORY_INDEXES = [
    'data.time',
    [('experiment_id', ASCENDING),
     ('data.time', ASCENDING),
     ('_id', ASCENDING)],
]
CONFIGURATION_INDEXES = [
    'experiment_id',
]
SECRETS_PATH = 'secrets.json'


# -- emitters -- #

DB_CONFIG_TYPE = {
    'connection_uri': 'string',
    'experiment_id': 'maybe[string]',
    'emit_limit': 'integer',
    'database': 'maybe[string]',
}


class DatabaseEmitter(Emitter):
    config_schema = {
        'connection_uri': 'string',
        'experiment_id': 'maybe[string]',
        'emit_limit': 'integer',
        'database': 'maybe[string]',
    }

    @abstractmethod
    def create_indexes(self, table: Any, columns: List[Any]) -> None:
        pass

    @property
    def current_pid(self):
        return os.getpid()

    def __init__(self, config=None, core=None):
        """Config may have 'host' and 'database' items. The config passed is expected to be:

                {'experiment_id':,
                 'emit_limit':,
                 'embed_path':}

                TODO: Automate this process for the user in builder
        """
        super().__init__(config, core)
        # self.core = core
        self.experiment_id = self.config.get('experiment_id', str(uuid.uuid4()))
        self.emit_limit = self.config.get('emit_limit', 4000000)


class MongoDatabaseEmitter(DatabaseEmitter):
    client_dict: Dict[int, MongoClient] = {}
    config_schema = {
        'connection_uri': 'string',
        'experiment_id': 'maybe[string]',
        'emit_limit': {
            '_type': 'integer',
            '_default': 4000000
        },
        'database': 'maybe[string]'
    }

    @classmethod
    def create_indexes(cls, table: Any, columns: List[Any]) -> None:
        """Create the listed column indexes for the given DB table."""
        for column in columns:
            table.create_index(column)

    def __init__(self, config, core) -> None:
        """Config may have 'host' and 'database' items. The config passed is expected to be:

                {'experiment_id':,
                 'emit_limit':,
                 'embed_path':}

                TODO: Automate this process for the user in builder
        """
        super().__init__(config)
        self.core = core
        self.experiment_id = self.config.get('experiment_id', str(uuid.uuid4()))
        # In the worst case, `breakdown_data` can underestimate the size of
        # data by a factor of 4: len(str(0)) == 1 but 0 is a 4-byte int.
        # Use 4 MB as the breakdown limit to stay under MongoDB's 16 MB limit.
        self.emit_limit = self.config['emit_limit']

        # create new MongoClient per OS process
        curr_pid = os.getpid()
        if curr_pid not in MongoDatabaseEmitter.client_dict:
            MongoDatabaseEmitter.client_dict[curr_pid] = MongoClient(
                config['connection_uri'])
        self.client: MongoClient = MongoDatabaseEmitter.client_dict[curr_pid]

        # extract objects from current mongo client instance
        self.db: Database = getattr(self.client, self.config.get('database', 'simulations'))
        self.history_collection: Collection = getattr(self.db, 'history')
        self.configuration: Collection = getattr(self.db, 'configuration')

        # create column indexes for the given collection objects
        self.create_indexes(self.history_collection, HISTORY_INDEXES)
        self.create_indexes(self.configuration, CONFIGURATION_INDEXES)

        self.fallback_serializer = make_fallback_serializer_function(self.core)

    def query(self, query):
        return self.history_collection.find_one(query)

    def history(self):
        return [v for v in self.history_collection.find()]

    def flush_history(self):
        for v in self.history():
            self.history_collection.delete_one(v)

    def update(self, inputs):
        self.history_collection.insert_one(inputs)
        return {}


def make_fallback_serializer_function(process_registry) -> Callable:
    """Creates a fallback function that is called by orjson on data of
    types that are not natively supported. Define and register instances of
    :py:class:`vivarium.core.registry.Serializer()` with serialization
    routines for the types in question."""

    def default(obj: Any) -> Any:
        # Try to lookup by exclusive type
        serializer = process_registry.access(str(type(obj)))
        if not serializer:
            compatible_serializers = []
            for serializer_name in process_registry.list():
                test_serializer = process_registry.access(serializer_name)
                # Subclasses with registered serializers will be caught here
                if isinstance(obj, test_serializer.python_type):
                    compatible_serializers.append(test_serializer)
            if len(compatible_serializers) > 1:
                raise TypeError(
                    f'Multiple serializers ({compatible_serializers}) found '
                    f'for {obj} of type {type(obj)}')
            if not compatible_serializers:
                raise TypeError(
                    f'No serializer found for {obj} of type {type(obj)}')
            serializer = compatible_serializers[0]
            if not isinstance(obj, Process):
                # We don't warn for processes because since their types
                # based on their subclasses, it's not possible to avoid
                # searching through the serializers.
                warn(
                    f'Searched through serializers to find {serializer} '
                    f'for data of type {type(obj)}. This is '
                    f'inefficient.')
        return serializer.serialize(obj)

    return default



from typing import *
import ast
from process_bigraph import Composite
from builder import Builder

from bsp.builders.builder import BSPBuilder


class BuildPrompter:
    """Front-End user interaction controller for high-level BioBuilder API."""
    builder_instance: Union[Builder, BSPBuilder] = None
    connect_all: bool = True

    @classmethod
    def generate_input_kwargs(cls, **config_params) -> Dict[str, Any]:
        """Generate kwargs to be used as dynamic input for process configuration.

            Args:
                **config_params:`kwargs`: values that would be otherwise defined by the user
                    in the prompter input prompt can instead be defined as kwargs in this
                    method. PLEASE NOTE: each kwarg value
                    passed here should be a dictionary specifying the process name (according
                    to the registry ie: `'CopasiProcess'`), and the process config kwarg.

                    For example:

                        prompter.add_single_process(
                            CopasiProcess={
                                'model': {
                                    'model_source': 'BIOMD0000000391'}

            Returns:
                Dict[str, Any]: configuration kwargs for process construction.
        """
        input_kwargs = {**config_params}
        process_kwargs = input('No config kwargs have yet been generated. Please enter the process configuration keyword arguments. Press enter to skip: ')
        if process_kwargs:
            process_args = process_kwargs.split(",")
            for i, arg in enumerate(process_args):
                arg = arg.strip()
                key, value = arg.split('=')
                try:
                    # safely evaluate the value to its actual data type
                    input_kwargs[key.strip()] = ast.literal_eval(value)
                except (ValueError, SyntaxError):
                    input_kwargs[key] = value
        print(f'Input kwargs generated: {input_kwargs}')
        return input_kwargs

    def add_single_process(self,
                           builder: Union[Builder, BSPBuilder] = None,
                           process_type: str = None,
                           config=None,
                           builder_node_name: str = None) -> None:
        """Get prompted through the steps of adding a single process to the bigraph via the builder."""
        if self.builder_instance is None:
            self.builder_instance = builder or BSPBuilder()

        if not process_type:
            process_type = input(
                f'Please enter one of the following process types that you wish to add:\n{self.builder_instance.list_processes()}\n:')

        if not builder_node_name:
            builder_node_name = input('Please enter the name that you wish to assign to this process: ')

        # generate input data from user prompt results and add processes to the bigraph  through pydantic model
        DynamicProcessConfig = self.builder_instance.get_pydantic_model(process_type)

        input_kwargs = self.generate_input_kwargs() if config is None else config.to_dict()
        dynamic_config = DynamicProcessConfig(**input_kwargs)
        self.builder_instance.add_process(
            process_id=builder_node_name,
            name=process_type,
            config=dynamic_config)  # {**input_kwargs})

        print(f'{builder_node_name} process successfully added to the bi-graph!')

        if self.connect_all:
            self.builder_instance.connect_all()
            print(f'All nodes including the most recently added {builder_node_name} processes connected!')

        print(f'Done adding single {builder_node_name} ({process_type}) to the bigraph.')
        return

    def add_processes(self,
                      num: int,
                      builder: Union[Builder, BSPBuilder] = None,
                      config=None,
                      write_doc: bool = False) -> None:
        """Get prompted for adding `num` processes to the bigraph and visualize the composite.

            Args:
                num:`int`: number of processes to add.
                builder:`Builder`: instance with which we add processes to bigraph
                config: used if not wanting to use input prompts. For now, this applies the same
                    config to each `num` of processes added. TODO: Expand this.
                write_doc: whether to write the doc. You will be re-prompted if False.

            # TODO: Allow for kwargs to be passed in place of input vals for process configs
        """
        print('Run request initiated...')
        print(f'{num} processes will be added to the bi-graph.')

        if self.connect_all:
            print('All processes will be connected as well.')
        else:
            # TODO: implement this through kwargs
            print('Using edge configuration spec...')

        for n in range(num):
            self.add_single_process(builder=builder, config=config)
        print('All processes added.')

        if not write_doc:
            write_doc = self.yesno(input('Save composition to document? (y/n): '))
        if write_doc:
            doc = self.builder_instance.document()
            doc_fp = input('Please enter the save destination of this document: ')
            self.builder_instance.write(filename=doc_fp)
            print('Composition written to document!')

        # view composite
        print('This is the composite: ')
        return self.visualize_bigraph()

    def generate_engine(self, builder: Builder = None) -> Composite:
        builder = builder or self.builder_instance
        return builder.generate()

    def generate_composite_run(self, duration: int = None, **run_params) -> None:
        """Generate and run a composite.

            Args:
                duration:`int`: the interval for which to run the composite.
        """
        if duration is None:
            duration = int(input('How long would you like to run this composite for?: '))

        print('Generating composite...')

        engine = self.generate_engine()
        print('Composite generated!')

        print(f'Running generated composite for an interval of {duration}')

        results = engine.run(duration)  # TODO: add handle force complete
        print('Composite successfully run. Request complete. Done.')
        return results

    def start(self, num: int = None):
        """Entrypoint to get prompted for input data with which to build the bigraph, then visualize
            and run the composite. All positional args and kwargs will be re-queried in the
            prompt if set to `None`. TODO: What other steps could possibly occur here? What about before?

            Args:
                num:`int`: number of processes to add. Defaults to `None`.
        """
        if num is None:
            num = int(input('How many processes would you like to add to the bigraph?'))
        return self.add_processes(num)

    def run(self, duration: int = None, **run_params) -> None:
        """Entrypoint to get prompted for input data with which to build the bigraph, then visualize
            and run the composite. All positional args and kwargs will be re-queried in the
            prompt if set to `None`. TODO: What other steps could possibly occur here? What about before?

            Args:
                duration:`int`: interval to run process composite for. Defaults to `None`.
                **run_params:`kwargs`: Custom params. TODO: implement these.
        """
        return self.generate_composite_run(duration=duration, **run_params)

    def execute(self, num: int = None, duration: int = None, **run_params) -> None:
        """For use as the highest level function called by the BioBuilder REST API."""
        self.start(num)
        return self.run(duration, **run_params)

    def write_bigraph(self, fp: str, output_dir='out'):
        """Use the builder instance attribute to write out the bigraph document to a fp."""
        return self.builder_instance.write(filename=fp, outdir=output_dir)

    def wipe_builder(self):
        self.builder_instance = None
        print(f'Builder flushed! This is verified because builder is: {self.builder_instance}')

    def flush(self, out_dir='out', fp: str = None, write: bool = True) -> None:
        """Optionally write the bigraph to document and wipe the current builder_instance instance.

            Args:
                out_dir:`str`: destination to which we save the document.
                fp:`str`: path by which to save if write is true. Defaults to None.
                write:`bool`: whether to write out the document before wiping. Defaults to `True`.
        """
        from datetime import datetime
        if write:
            if fp is None:
                fp = f"BUILD_FLUSH_{str(datetime.now()).replace(' ', '__').replace(':', '_')}"
            self.write_bigraph(fp, out_dir)
        return self.wipe_builder()

    def visualize_bigraph(self):
        return self.builder_instance.visualize()

    @classmethod
    def yesno(cls, user_input: str) -> Union[bool, None]:
        return True if 'y' in user_input.lower() \
                else False if 'n' in user_input.lower() \
                else None






from bsp.schemas import config, types
from bsp.registration import Registrar
from bsp.implementations import INITIAL_MODULES


VERBOSE_REGISTRATION = True
ATTEMPT_INSTALL = False


# project-scoped process/implementation registrar
app_registrar = Registrar()


#  register types:
app_registrar.register_initial_types(
    config=config,
    types=types,
    verbose=VERBOSE_REGISTRATION
)


# register implementations of steps and processes:
app_registrar.register_initial_modules(
    items_to_register=INITIAL_MODULES,
    verbose=VERBOSE_REGISTRATION,
    attempt_install=ATTEMPT_INSTALL
)












"""
Data model relating to quantum data in Qiskit and Cirq.
"""


from bsp.data_model.base import BaseClass
import numpy as np
from dataclasses import dataclass


class DiracNotation(np.ndarray):
    def __new__(cls, values: list[complex]):
        return np.asarray(values).view(cls)


class Bra(DiracNotation):
    pass


class Ket(DiracNotation):
    def bra(self) -> Bra:
        ket_value = self.view()
        return Bra(np.conj(ket_value).T)


class XYZMolecule(str):
    def __new__(cls, *args, **kwargs):
        return str.__new__(cls, *args, **kwargs)

    @property
    def elements(self, unique: bool = False) -> list:
        items = self.split()
        elements = [line[0] for line in [ln.strip() for ln in self.split(";")]]
        return list(set(elements)) if unique else elements


class MoleculeAtom:
    pass





from enum import Enum


class CompatibleUtcSimulators(Enum):
    AMICI = ('amici', '0.11.21')
    COPASI = ('copasi', '0.71')
    PYSCES = ('pysces', '1.2.2')
    TELLURIUM = ('tellurium', '0.85')


class CompatibleStochasticParticleSimulators(Enum):
    READDY = 'readdy'
    SMOLDYN = 'smoldyn'


COMPATIBLE_UTC_SIMULATORS = [CompatibleUtcSimulators[member].value for member in list(dict(CompatibleUtcSimulators.__members__).keys())]


"""
Membrane process using forward euler gradient descent integration.
"""

import inspect
import os
import shutil
from functools import partial
from pathlib import Path
from typing import Dict, Union, List, Tuple
import tempfile as tmp

import numpy as np
import pymem3dg as dg
import pymem3dg.boilerplate as dgb
from netCDF4 import Dataset
from process_bigraph import Process, ProcessTypes

from bsp.utils.membrane_utils import extract_data, parse_ply, calculate_preferred_area, new_parameters, extract_last_data


class SimpleMembraneProcess(Process):
    config_schema = {
        'mesh_file': 'MeshFileConfig',
        'geometry': 'GeometryConfig',
        'tension_model': 'TensionModelConfig',
        'osmotic_model': 'OsmoticModelConfig',
        'parameters': 'ParametersConfig',
        'save_period': 'integer',
        'tolerance': 'float',
        'characteristic_time_step': 'integer',
        'console_output': 'boolean',
    }

    def __init__(self, config: Dict[str, Union[Dict[str, float], str]] = None, core=None):
        super().__init__(config, core)
        # get simulation params
        self.save_period = self.config.get("save_period", 1)  # interval at which sim state is saved to disk/recorded
        self.tolerance = self.config.get("tolerance", 1e-11)
        self.characteristic_time_step = self.config.get("characteristic_time_step", 0.02)
        self.tension_modulus = self.config["tension_model"].get("modulus", 0.1)
        self.console_output = self.config.get("console_output", True)

        # parse input of either mesh file or geometry spec
        initial_faces = None
        initial_vertices = None

        mesh_file = self.config.get("mesh_file")
        if mesh_file:
            geometry = dg.Geometry(mesh_file)
            initial_faces = geometry.getFaceMatrix()
            initial_vertices = geometry.getVertexMatrix()
        else:
            geometry = self.config.get("geometry")
            shape = geometry['type']
            mesh_constructor = getattr(dg, f'get{shape.replace(shape[0], shape[0].upper())}')
            initial_faces, initial_vertices = mesh_constructor(**geometry['parameters'])

        # this geometry is initially parameterized by the geometry spec, but remains stateful as it mutates during update calls
        self.geometry = dg.Geometry(initial_faces, initial_vertices)
        self.n_vertices = self.geometry.getVertexMatrix().shape[0]

        # get parameters, osmotic, and tension params from config
        self.param_spec = self.config.get("parameters")
        self.parameters = new_parameters(self.param_spec)

        self.system = dg.System(
            geometry=self.geometry,
            parameters=self.parameters,
        )

        self.osmotic_model_spec = self.config.get("osmotic_model")
        self.tension_model_spec = self.config.get("tension_model")
        self.iterations = 0

    def initial_state(self):
        initial_face_matrix = self.geometry.getFaceMatrix().tolist()
        initial_vertices = self.geometry.getVertexMatrix()
        initial_geometry = {
            "faces": initial_face_matrix,
            "vertices": initial_vertices.tolist(),
        }

        # set initial velocities to an array of the correct shape to 0.0, 0.0, 0.0 TODO: should this be different?
        self.system.initialize(True)
        initial_velocities = self.system.getVelocity().tolist()  # np.zeros(initial_vertices.shape).tolist()

        # set initial protein density (gradient) to constant (1.) for all vertices
        initial_protein_density = self.system.getProteinDensity().tolist()  # np.ones(initial_vertices.shape[0]).tolist()

        # set initial volume and surface area from config
        initial_volume = self.osmotic_model_spec["volume"]
        initial_preferred_volume = self.osmotic_model_spec["preferred_volume"]
        initial_res_volume = self.osmotic_model_spec["reservoir_volume"]
        initial_surface_area = self.geometry.getSurfaceArea()

        # similarly set no forces as initial output
        initial_net_forces = np.zeros(initial_vertices.shape).tolist()
        initial_notable_vertices = self.geometry.getNotableVertex()

        return {
            "geometry": initial_geometry,
            "velocities": initial_velocities,
            "protein_density": initial_protein_density,
            "volume": initial_volume,
            "preferred_volume": initial_preferred_volume,
            "reservoir_volume": initial_res_volume,
            "surface_area": initial_surface_area,
            "net_forces": initial_net_forces,
            "notable_vertices": initial_notable_vertices
        }

    def inputs(self):
        """
        - The first three types are used to parameterize the kth Geometry
        - The preferred_volume, current_volume, and osmotic_strength is used to parameterize the kth Osmotic pressure model.
        - The tension model is parameterized by inference into the osmotic model. The modulus is static.
        """
        return {
            'geometry': 'GeometryType',  # 'tree',
            'protein_density': 'ProteinDensityType',  # 'list',
            'velocities': 'VelocitiesType',  # 'list',
            'volume': 'float',
            'preferred_volume': 'float',
            'reservoir_volume': 'float',
            'surface_area': 'float',
            'osmotic_strength': 'float',
        }

    def outputs(self):
        return {
            'geometry': 'GeometryType',
            'protein_density': 'ProteinDensityType',
            'velocities': 'VelocitiesType',
            'volume': 'float',
            'preferred_volume': 'float',
            'reservoir_volume': 'float',
            'surface_area': 'float',
            'net_forces': 'MechanicalForcesType',
            'notable_vertices': 'integer',
        }

    def update(self, state, interval):
        # parameterize kth geometry from k-1th outputs
        previous_vertices = np.array(state["geometry"]["vertices"][:self.n_vertices])  # TODO: why is -1 in improper format?
        self.geometry.setInputVertexPositions(previous_vertices)

        # set the kth osmotic volume model  # dfba vals here in update
        preferred_volume_k = state["preferred_volume"]  # TODO: should this be constant/static?
        reservoir_volume_k = state["reservoir_volume"]
        osmotic_strength_k = self.osmotic_model_spec["strength"]
        osmotic_model_k = partial(
            dgb.preferredVolumeOsmoticPressureModel,
            preferredVolume=preferred_volume_k,  # make input port here if value has changed (fba)
            reservoirVolume=reservoir_volume_k,  # output port
            strength=osmotic_strength_k,
        )

        # set the surface area tension model
        preferred_area_k = calculate_preferred_area(v_preferred=preferred_volume_k)
        tension_model_k = partial(
            dgb.preferredAreaSurfaceTensionModel,
            modulus=self.tension_modulus,
            # preferredArea=self.geometry.getSurfaceArea(),
            preferredArea=preferred_area_k,  # self.tension_model_spec["preferred_area"],
        )

        # update the tension and osmotic params
        parameters = new_parameters(self.param_spec)
        parameters.tension.form = tension_model_k
        parameters.osmotic.form = osmotic_model_k

        # instantiate and initialize kth system using Geometry, proteinDensity, velocity, and parameters
        protein_density_k = np.array(state["protein_density"])
        velocities_k = np.array(state["velocities"])

        # we parameterize the kth system with the stateful geometry that has been updated with the latest vertex coordinates
        # system_k = dg.System(
        #     geometry=self.geometry,
        #     proteinDensity=protein_density_k,
        #     velocity=velocities_k,
        #     parameters=self.parameters,
        # )
        self.system.parameters = parameters
        # self.system.initialize(True)
        self.system.updateConfigurations()
        # system_k.initialize()

        duration_k = interval + state['duration']# self.iterations + 1

        # set up solver and parse time params
        output_dir_k = Path(tmp.mkdtemp())
        integrator_k = dg.VelocityVerlet(
            system=self.system,
            characteristicTimeStep=self.characteristic_time_step,
            totalTime=interval,
            savePeriod=interval,
            tolerance=self.tolerance,
            outputDirectory=str(output_dir_k)
        )
        integrator_k.ifPrintToConsole = self.console_output
        integrator_k.ifOutputTrajFile = True

        # run integration
        success = integrator_k.integrate()  # or should this be integrator_k.step(interval)?

        # uncomment and implement below for IO style update
        # output_path_k = str(output_dir_k / "traj.nc")
        # data = Dataset(output_path_k, 'r')

        # get kth protein densities from kth system
        output_protein_density = self.system.getProteinDensity().tolist()

        # get kth velocities from kth system
        output_velocities = self.system.getVelocity().tolist()

        # verify geometry instance by getting it from the system (in case it has mutated). TODO: is this needed?
        output_geometry = {
            "faces": self.geometry.getFaceMatrix().tolist(),
            "vertices": self.geometry.getVertexMatrix().tolist(),
        }

        # get kth volume and surface area from output geometry
        vol_variation_vectors = self.geometry.getVertexVolumeVariationVectors()  # TODO: this is not yet used
        output_volume = self.geometry.getVolume()
        output_surface_area = self.geometry.getSurfaceArea()

        # get kth mechanical force vectors (that is, the net sum of x, y, and z vectors for each vertex)
        forces_k = self.system.getForces()
        output_force_vectors = forces_k.getMechanicalForceVec().tolist()

        notable_vertices = self.geometry.getNotableVertex()

        # clean up temporary files
        shutil.rmtree(str(output_dir_k))
        print(f'Duration at interval {interval}: {interval}')
        self.iterations += 1
        # TODO: refactor geometry to include not. vert.
        return {
            'geometry': output_geometry,
            'protein_density': output_protein_density,
            'velocities': output_velocities,
            'volume': output_volume,
            'preferred_volume': preferred_volume_k,
            'reservoir_volume': reservoir_volume_k,
            'surface_area': output_surface_area,
            'net_forces': output_force_vectors,
            'notable_vertices': notable_vertices
        }



import os
from tests import ProcessUnitTest


def test_process():
    smoldyn_models_root = 'biosimulators_processes/model_files/smoldyn'
    model_name = 'minE_model.txt'
    model_fp = os.path.join(smoldyn_models_root, model_name)

    instance = {
        'smoldyn': {
            '_type': 'process',
            'address': 'local:smoldyn',
            'config': {
                'model_filepath': model_fp,
            },
            'inputs': {
                'species_counts': ['species_counts_store'],
                'molecules': ['molecules_store']},
            'outputs': {
                'species_counts': ['species_counts_store'],
                'molecules': ['molecules_store']}
        },
        # 'emitter': {
        #     '_type': 'step',
        #     'address': 'local:ram-emitter',
        #     'config': {
        #         'emit': {
        #             'species_counts': 'tree[float]',
        #             'molecules': 'tree[float]',
        #         },
        #     },
        #     'inputs': {
        #         'floating_species': ['species_counts_store'],
        #         'molecules': ['molecules_store'],
        #     }
        # }
    }

    return ProcessUnitTest(instance_doc=instance, duration=10)


test_process()



import numpy as np
import networkx as nx
from qiskit.primitives import Sampler
from qiskit_algorithms.optimizers import COBYLA
from qiskit_algorithms.utils import algorithm_globals
from qiskit_algorithms import QAOA as QAOASolver
from process_bigraph import Process

from bsp.utils.quantum_utils import get_operator, sample_most_likely


class QAOA(Process):
    config_schema = {
        "n_variables": "integer",
        "random_seed": "integer"
        # "edge_list": "list[tuple[float]]",  # in the format: [(node_a, node_b, weight/is_connected(1 or 0)), ...]
    }

    def __init__(self, config, core):
        super().__init__(config, core)
        self.n_variables = self.config["n_variables"]
        self.random_seed = self.config.get("random_seed", 10598)

        # initial params for probablistic terms
        self.initial_gamma = np.pi
        self.initial_beta = np.pi/2
        self.init_params = [self.initial_gamma, self.initial_beta, self.initial_gamma, self.initial_beta]

    def initial_state(self):
        return {
            "bitstring": [0 for _ in range(self.n_variables)],
            "n_nodes": self.n_variables
        }

    def inputs(self):
        return {
            "n_nodes": "integer",
            "adjacentcy_matrix": "list[list[float]]"
        }

    def outputs(self):
        return {
            "bitstring": "list[integer]",
            "n_nodes": "integer"
        }

    def update(self, inputs, interval):
        # graph = initialize_graph_k(n_nodes_k)

        # initialize graph
        n_nodes_k = inputs.get("n_nodes")
        w = np.array(
            inputs.get("adjacentcy_matrix")
        )
        G = nx.from_numpy_array(w)

        # get quantum operator
        qubit_op, offset = get_operator(w, n_nodes_k)

        # set up optimizer and sampler
        optimizer = COBYLA()
        sampler = Sampler()
        algorithm_globals.random_seed = self.random_seed

        # perform qaoa
        qaoa = QAOASolver(sampler, optimizer, reps=2)

        # extract bitstring
        result = qaoa.compute_minimum_eigenvalue(qubit_op)
        bitstring_k = sample_most_likely(result.eigenstate)

        return {
            "bitstring": bitstring_k.tolist(),
            "n_nodes": n_nodes_k - 1
        }



from abc import ABC, abstractmethod

from process_bigraph import Process


class SedProcess(Process, ABC):
    config_schema = {
        'model': 'sed_model',
        'species_context': {
            '_type': 'string',
            '_default': 'concentrations'
        },
        'method': {
            '_type': 'string',
            '_default': 'deterministic'  # <-- CVODE for consistency, or should we use LSODA?
        },
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

    @abstractmethod
    def initial_state(self):
        pass

    @abstractmethod
    def inputs(self):
        pass

    @abstractmethod
    def outputs(self):
        pass

    @abstractmethod
    def update(self, inputs, interval):
        pass


class SedUTCProcess(SedProcess):
    config_schema = {
        'model': 'TimeCourseConfig',
        'species_context': {
            '_type': 'string',
            '_default': 'concentrations'
        },
        'method': {
            '_type': 'string',
            '_default': 'deterministic'  # <-- CVODE for consistency, or should we use LSODA?
        },
        'sbml_fp': 'maybe[string]'
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

    @abstractmethod
    def initial_state(self):
        pass

    @abstractmethod
    def inputs(self):
        pass

    @abstractmethod
    def outputs(self):
        pass

    @abstractmethod
    def update(self, inputs, interval):
        pass



import os

from process_bigraph import Composite, pp
from biosimulators_processes import CORE
from bsp.utils.helpers import prepare_single_ode_process_document


def test_process():
    biomodel_id = 'BIOMD0000000630'
    model_fp = f'biosimulator_processes/model_files/sbml/BIOMD0000000630_url.xml'
    species_context = 'counts'
    species_port_name = f'floating_species_{species_context}'
    species_store = [f'floating_species_{species_context}_store']
    duration = 10

    print(os.path.exists(model_fp))

    instance = prepare_single_ode_process_document(
        simulator_name='copasi',
        process_id=f'{biomodel_id}_copasi_process',
        sbml_model_fp=model_fp)

    composition = Composite(config={'state': instance}, core=CORE)
    composition.run(duration)
    results = composition.gather_results()
    return pp(results)


# def test_process_from_document():
#     # read the document from local file:
#     five_process_fp = 'composer-notebooks/out/five_process_composite.json'
#     with open(five_process_fp, 'r') as fp:
#         instance = json.load(fp)
#
#     return ProcessUnitTest(instance, duration=10)


if __name__ == '__main__':
    test_process()



"""
Type schema definitions relating to process implementations within this application.

The type name suffix is used to denote its function. Currently there are two suffixes:

- ...Config -> this relates to the schema for configuration data only (constructor)
"""


__all__ = [
    'DatabaseConfig',
    'GeometryConfig',
    'MeshFileConfig',
    'ParametersConfig',
    'OsmoticModelConfig',
    'SBMLFileCobraConfig',
    'SBMLModelChangesConfig',
    'SBMLModelConfig',
    'TimeCourseConfig',
    'TensionModelConfig',
    'SedModelConfig'
]


def check_mesh_file(state, schema, core):
    # TODO: check the state for .ply extension in filepath within config
    pass


# -- configs: that is, schemas related to config_schema specs only --

DatabaseConfig = {
    'connection_uri': 'string',
    'experiment_id': 'maybe[string]',
    'emit_limit': 'integer',
    'database': 'maybe[string]',
}

GeometryConfig = {
    'type': 'string',  # add enum[..] if used, ie; 'icosphere'
    'parameters': 'tree'  # TODO:  params required for aforementioned shape type
}

MeshFileConfig = {
    '_inherit': 'string',
    '_check': check_mesh_file
}

# TODO: fully implement this with all attributes
ParametersConfig = 'tree'

SBMLFileCobraConfig = {
    '_inherit': 'string',
    # '_check': check_sbml,
    '_apply': 'set',
}

SBMLModelChangesConfig = {
    'species_changes': {
        'species_name': {
            'unit': 'maybe[string]',
            'initial_concentration': 'maybe[float]',
            'initial_particle_number': 'maybe[float]',
            'initial_expression': 'maybe[string]',
            'expression': 'maybe[string]'
        }
    },
    'global_parameter_changes': {
        'global_parameter_name': {
            'initial_value': 'maybe[float]',
            'initial_expression': 'maybe[string]',
            'expression': 'maybe[string]',
            'status': 'maybe[string]',
            'type': 'maybe[string]'  # (ie: fixed, assignment, reactions)
        }
    },
    'reaction_changes': {
        'reaction_name': {
            'parameters': {
                'reaction_parameter_name': 'maybe[int]'  # (new reaction_parameter_name value)  <-- this is done with set_reaction_parameters(name="(REACTION_NAME).REACTION_NAME_PARAM", value=VALUE)
            },
            'reaction_scheme': 'maybe[string]'   # <-- this is done like set_reaction(name = 'R1', scheme = 'S + E + F = ES')
        }
    }
}

SedModelConfig = {
    'model_id': 'string',
    'model_source': 'string',
    'model_language': {
        '_default': 'string',
        '_type': 'string'
    },
    'model_name': 'string',
    'model_units': 'tree',
    'model_changes': 'tree'
}

SBMLModelConfig = {
    'model_id': 'string',
    'model_source': 'string',
    'model_language': {
        '_default': 'string',
        '_type': 'string'
    },
    'model_name': 'string',
    'model_units': 'tree[string]',
    'model_changes': SBMLModelChangesConfig  # 'tree[string]'
}

TimeCourseConfig = {
    'model': SBMLModelConfig,
    'time_config': 'tree[string]',  # ie: {start, stop, steps}
    'species_context': {
        '_default': 'concentrations',
        '_type': 'string'
    },
    'working_dir': 'string'
}

TensionModelConfig = {
    'modulus': 'float',
    'preferredArea': 'float'
}

OsmoticModelConfig = {
    'volume': 'float',
    'preferred_volume': 'float',
    'reservoir_volume': 'float',
    'strength': 'float'  # what units is this in??
}




import os
import logging
from tempfile import mkdtemp

import libsbml
import numpy as np
import seaborn as sns
from amici import amici, SbmlImporter, import_model_module, runAmiciSimulation
from matplotlib import pyplot as plt
from process_bigraph import Step

from bsp.data_model.sed import UTC_CONFIG_TYPE
from bsp.utils.helpers import calc_duration, calc_num_steps, calc_step_size, check_ode_kisao_term


class UtcAmici(Step):
    """
       Parameters:
            config:`Dict`: dict keys include:
                model: SED Model Spec.
                species_context: Context by which to measure species outputs (defaults to concentrations).
                model_output_dir: Dirpath in which to save the AMICI model output.
                observables: for example:
                    observables = {
                        "observable_x1": {"name": "", "formula": "x1"},
                        "observable_x2": {"name": "", "formula": "x2"},
                        "observable_x3": {"name": "", "formula": "x3"},
                        "observable_x1_scaled": {"name": "", "formula": "scaling_x1 * x1"},
                        "observable_x2_offsetted": {"name": "", "formula": "offset_x2 + x2"},
                        "observable_x1withsigma": {"name": "", "formula": "x1"},
                    }
                constant_parameters: for example:
                    constant_parameters = ["k0"]
                sigmas: for example:
                    sigmas = {"observable_x1withsigma": "observable_x1withsigma_sigma"}

    """
    config_schema = UTC_CONFIG_TYPE

    # AMICI-specific fields
    config_schema['model_output_dir'] = {
        '_default': mkdtemp(),
        '_type': 'string'
    }
    config_schema['observables'] = 'maybe[tree[string]]'
    config_schema['constant_parameters'] = 'maybe[list[string]]'
    config_schema['sigmas'] = 'maybe[tree[string]]'

    def __init__(self,
                 config=None,
                 core=None,
                 time_config: dict = None,
                 model_source: str = None,
                 sed_model_config: dict = None):
        self.slice = ()
        # A. no config but either an omex file/dir or sbml file path
        configuration = config or {}
        source = configuration.get('model').get('model_source')
        archive_dir_source = os.path.isdir(source)
        if not configuration and model_source:
            configuration = {'model': {'model_source': model_source}}

        # B. has a config but wishes to override TODO: fix this.
        if sed_model_config and configuration:
            configuration['model'] = sed_model_config

        # C. has a source of sbml file path as expected.
        elif source and source.endswith('.xml') and not source.lower().startswith('manifest'):
            pass

        # D. has a config passed with an archive dirpath or filepath or sbml filepath as its model source:
        else:
            omex_path = configuration.get('model').get('model_source')
            # Da: user has passed a dirpath of omex archive or the path to an unzipped archive as model source
            archive_dir = unpack_omex_archive(archive_filepath=source, working_dir=config.get('working_dir') or mkdtemp()) if not archive_dir_source else source

            # set expected model path for init
            configuration['model']['model_source'] = get_archive_model_filepath(archive_dir)

            # extract the time config from archive's sedml
            configuration['time_config'] = self._get_sedml_time_params(archive_dir)

        if time_config and not len(configuration.get('time_config', {}).keys()):
            configuration['time_config'] = time_config

        super().__init__(config=configuration, core=core)

        # reference model source and assert filepath
        model_fp = self.config['model'].get('model_source')
        assert model_fp is not None and '/' in model_fp, 'You must pass a valid path to an SBML model file.'
        model_config = self.config['model']

        # get and compile libsbml model from fp
        sbml_reader = libsbml.SBMLReader()
        sbml_doc = sbml_reader.readSBML(model_fp)
        self.sbml_model_object: libsbml.Model = sbml_doc.getModel()

        # get model args from config
        model_id = self.config['model'].get('model_id', None) \
            or model_fp.split('/')[-1].replace('.', '_').split('_')[0]

        model_output_dir = self.config['model_output_dir']

        # compile sbml to amici
        sbml_importer = SbmlImporter(model_fp)
        sbml_importer.sbml2amici(
            model_name=model_id,
            output_dir=model_output_dir,
            verbose=logging.INFO,
            observables=self.config.get('observables'),
            constant_parameters=self.config.get('constant_parameters'),
            sigmas=self.config.get('sigmas'))
        model_module = import_model_module(model_id, model_output_dir)
        self.amici_model_object: amici.Model = model_module.getModel()

        # set species context (concentrations for ODE by default)
        self.context_type = self.config['species_context']
        self.species_context_key = f'floating_species'
        self.use_counts = 'counts' in self.species_context_key

        # get species names
        self.floating_species_list = list(self.amici_model_object.getStateIds())
        self.floating_species_initial = list(self.amici_model_object.getInitialStates())
        self.sbml_species_ids = [spec for spec in self.sbml_model_object.getListOfSpecies()]
        self.sbml_species_mapping = dict(zip(
            list(map(lambda s: s.name, self.sbml_species_ids)),
            [spec.getId() for spec in self.sbml_species_ids],
        ))

        # get model parameters
        self.model_parameter_objects = self.sbml_model_object.getListOfParameters()
        self.model_parameters_list = [param.getName() for param in self.model_parameter_objects]
        self.model_parameters_values = [param.getValue() for param in self.model_parameter_objects]

        # get reactions
        self.reaction_objects = self.sbml_model_object.reactions
        self.reaction_list = [reaction.getName() for reaction in self.reaction_objects]

        # get method
        self.method = self.amici_model_object.getSolver()

        # set time config and model with time config
        utc_config = self.config.get('time_config')
        assert utc_config, \
            "For now you must manually pass time_config: {duration: , num_steps: , step_size: , } in the config."
        self.step_size = utc_config.get('step_size')
        self.duration = utc_config.get('duration')
        self.num_steps = utc_config.get('num_steps')
        self.initial_time = utc_config.get('initial_time') or 0
        self.output_start_time = utc_config.get('output_start_time')

        if len(list(utc_config.keys())) < 3:
            self._set_time_params()

        self.t = np.linspace(self.output_start_time, self.duration, self.num_steps)

        self.amici_model_object.setTimepoints(self.t)
        self._results = {}
        self.output_keys = [list(self.sbml_species_mapping.keys())[i] for i, spec_id in enumerate(self.floating_species_list)]

    def initial_state(self):
        species_initial = dict(zip(self.floating_species_list, self.floating_species_initial))
        model_params_initial = dict(zip(self.model_parameters_list, self.model_parameters_values))
        return {
            'time': [0.0],
            'floating_species': species_initial,
            'model_parameters': model_params_initial,
            'reactions': self.reaction_list
        }

    @staticmethod
    def _get_sedml_time_params(omex_path: str):
        sedml_fp = None
        for f in os.listdir(omex_path):
            if f.endswith('.sedml'):
                sedml_fp = os.path.join(omex_path, f)

        assert sedml_fp is not None, 'Your OMEX archive must contain a valid SEDML file.'
        # sedml_fp = os.path.join(omex_path, 'simulation.sedml')
        sedml_utc_config = get_sedml_time_config(sedml_fp)

        def convert(x):
            return int(x) if isinstance(x, float) else int(float(x))

        output_end = convert(sedml_utc_config['outputEndTime'])
        output_start = convert(sedml_utc_config['outputStartTime'])
        duration = output_end - output_start
        n_steps = convert(sedml_utc_config['numberOfPoints'])
        initial_time = convert(sedml_utc_config['initialTime'])
        step_size = calc_step_size(duration, n_steps)

        # check kisao id for supported algorithm/kisao ID
        specified_alg = sedml_utc_config['algorithm'].split(':')[1]
        supported_alg = check_ode_kisao_term(specified_alg)
        if not supported_alg:
            raise Exception('Algorithm specified in OMEX archive is non-deterministic and thus not supported by a Uniform Time Course implementation.')

        return {
            'duration': output_end,  # duration,
            'num_steps': n_steps,  # to account for self comparison
            'step_size': step_size,
            'output_start_time': output_start,
            'initial_time': initial_time
        }

    def plot_results(self, flush=True):
        """Plot ODE simulation observables with Seaborn."""
        plt.figure(figsize=(20, 8))
        for n in range(len(self.floating_species_list)):
            sns.lineplot(x=self._results['time'], y=list(self._results['floating_species'].values())[n])
        plt.xlabel('Time')
        plt.ylabel('Concentrations')
        plt.title('Species Concentrations over Time with AMICI')
        return self.flush_results() if flush else None

    def flush_results(self):
        return self._results.clear()

    def _set_time_params(self):
        if self.step_size and self.num_steps:
            self.duration = calc_duration(self.num_steps, self.step_size)
        elif self.step_size and self.duration:
            self.num_steps = calc_num_steps(self.duration, self.step_size)
        else:
            self.step_size = calc_step_size(self.duration, self.num_steps)

    def inputs(self):
        # dependent on species context set in self.config
        return {
            'time': 'list[float]',
            self.species_context_key: 'tree[float]',
            'model_parameters': 'tree[float]',
            'reactions': 'list[string]'}

    def outputs(self):
        return {
            'time': 'list[float]',
            self.species_context_key: 'tree[float]'}  # floating_species_type}

    def _generate_results(self, inputs=None):
        x = inputs or self.initial_state()
        if len(x.keys()):
            set_values = []
            for species_id, value in x[self.species_context_key].items():
                set_values.append(value)
            self.amici_model_object.setInitialStates(set_values)

        result_data = runAmiciSimulation(solver=self.method, model=self.amici_model_object)

        # TODO: ensure that `keys` are threadsafe.
        floating_species_results = dict(zip(
            self.output_keys,
            list(map(lambda x: result_data.by_id(f'{x}'), self.floating_species_list))))

        return {
            'time': self.t,
            self.species_context_key: floating_species_results}

    def update(self, inputs=None):
        results = self._generate_results(inputs)
        self._results = results.copy()
        return results



from dataclasses import dataclass
from typing import *

from process_bigraph import Composite, ProcessTypes
from builder import Builder

from bsp.data_model.base import BaseClass
from bsp.registration import Registrar


class BSPBuilder(Builder):
    _is_initialized = False

    def __init__(self, schema: Dict = None, tree: Dict = None, filepath: str = None, core: ProcessTypes = None):
        if not self._is_initialized:
            super().__init__(schema=schema, tree=tree, file_path=filepath, core=core)


@dataclass
class CompositionComponent(BaseClass):
    name: str

    def to_dict(self):
        node_dict = super().to_dict()
        node_dict.pop("name")
        return node_dict


@dataclass
class CompositionPort(CompositionComponent):
    store: List[str]


@dataclass
class CompositionNode(CompositionComponent):
    _type: str
    address: str
    config: Dict[str, Any]
    inputs: Dict[str, List[str]]
    outputs: Dict[str, List[str]]


class CompositionDocument(dict):
    pass


class CompositionBuilder:
    nodes: List[CompositionNode]
    document: CompositionDocument

    def __init__(self,
                 registrar: Registrar,
                 nodes: Optional[List[CompositionNode]] = None):
        self.registrar = registrar
        self.nodes = nodes
        self.document = self.to_document()

    def to_document(self) -> CompositionDocument:
        import warnings

        doc = {}
        if len(self.nodes):
            for node in self.nodes:
                doc[node.name] = node.to_dict()
        else:
            warnings.warn("There are no nodes in the composition")

        return CompositionDocument(doc)

    def generate_composite(self) -> Composite:
        doc = self.to_document()
        core = self.registrar.core

        return Composite({
            "state": doc,
            "emitter": {"mode": "all"}
        }, core=core)

    def add_node(self, node: CompositionNode):
        self.nodes.append(node)
        self.document = self.to_document()

    def remove_node(self, node_name: str):
        if node_name not in list(self.document.keys()):
            raise ValueError(f"{node_name} not found in document.")
        else:
            for node in self.nodes:
                if node.name == node_name:
                    self.nodes.remove(node)
            self.document.pop(node_name)


def test(registrar: Optional[Registrar] = None):
    qaoa_spec = CompositionNode(
        name="qaoa",
        _type="process",
        address="local:qaoa",
        config={
            "n_variables": 5
        },
        inputs={
            "n_nodes": ["n_nodes_store"],
            "adjacentcy_matrix": ["adjacentcy_matrix_store"]
        },
        outputs={
            "n_nodes": ["n_nodes_store"],
            "bitstring": ["bitstring_store"]
        }
    )

    builder = CompositionBuilder(registrar, [qaoa_spec])
    doc = builder.to_document()

    doc['emitter'] = {
        "_type": "step",
        "address": "local:ram-emitter",
        "config": {
            "emit": {
                "n_nodes": "integer",
                "bitstring": "list[integer]"
            }
        },
        "inputs": {
            'n_nodes': ['n_nodes_store'],
            'bitstring': ['bitstring_store']
        }
    }



import os
import json

from process_bigraph import Composite, pp
from biosimulators_processes import CORE
from biosimulators_processes.data_model.compare_data_model import ODEComparisonDocument


def test_ode_comparison_process():
    biomodel_id = 'BIOMD0000000630'
    model_fp = f'biosimulator_processes/model_files/sbml/BIOMD0000000630_url.xml'
    duration = 30

    comparison_instance = ODEComparisonDocument(
        simulators={''}
    )



"""As of 05/18/2024, most of the process implementations are considered Sed Processes"""

import os

from process_bigraph import Composite

from biosimulators_processes import CORE
from bsp.utils.helpers import prepare_single_ode_process_document
from biosimulators_processes.data_model.compare_data_model import ODECompositionResult


def test_ode_sed_processes(model_entrypoint: str, num_steps: int, duration: int = 1):
    # TODO: for now, just testing ode processes. Expand this *asap.
    local_ode_sed_process_addresses = ['tellurium', 'copasi', 'amici']
    return [
        ODECompositionResult(duration=duration, num_steps=num_steps, model_entrypoint=model_entrypoint, simulator_names=[sim_name])
        for sim_name in local_ode_sed_process_addresses]


def test_single_ode_sed_process(process_address: str, biomodel_id: str = None, model_fp: str = None):
    # biomodel_id = 'BIOMD0000000630'
    model_source = model_fp or biomodel_id
    assert model_source is not None, "You must provide either a biomodel id or an sbml model filepath"
    species_context = 'counts'
    species_port_name = f'floating_species_{species_context}'
    species_store = [f'floating_species_{species_context}_store']
    duration = 10

    assert os.path.exists(model_fp), "You must provide a valid sbml model file path"

    process_id = f'{model_fp.split("/")[-1] if model_fp else model_source}_test_process'
    instance = prepare_single_ode_process_document(
        simulator_name=process_address,
        process_id=process_id,
        sbml_model_fp=model_fp)

    composition = Composite(config={'state': instance}, core=CORE)
    composition.run(duration)
    return composition.gather_results()



from typing import *
from abc import abstractmethod

import numpy as np
from process_bigraph import Step
from biosimulators_processes.processes.copasi_process import CopasiProcess


class ParameterScan(Step):
    """
        For example:

            parameters = {
                'global': {
                    'ADP': <STARTING_VALUE>,
                    ...},
                'species': {
                    'species_a': <STARTING_VALUE>,
                    'species_b': ...
                    ...},
                'reactions': {
                    'R1': {
                        '(R1).k1': <STARTING_VALUE>,
                    ...}
            }

            ...according to the config schema for Copasi Process (model changes)


    """
    config_schema = {
        'n_iterations': 'int',
        'perturbation_magnitude': 'float',
        'parameters': 'tree[string]',  # ie: 'param_type: param_name'... like 'global': 'ADP', 'species': 'S3'
        'process_config': 'tree[string]'}  # use the builder to extract this

    @abstractmethod
    def inputs(self):
        pass

    @abstractmethod
    def outputs(self):
        pass

    @abstractmethod
    def initial_state(self):
        pass

    @abstractmethod
    def update(self, input):
        pass


class DeterministicTimeCourseParameterScan(Step):
    """Using CopasiProcess as the primary TimeCourse simulator.

        # TODO: enable multiple Simulator types.

        should at some point in the stack return a list or dict of configs that can be used by CopasiProcess which are then
            used to run during update
    """
    config_schema = {
        'process_config': 'tree[string]',
        'n_iterations': 'int',
        'iter_stop': 'float',
        'iter_start': 'maybe[float]',
        'perturbation_magnitude': 'float',
        'parameters': 'list[object]'}

    def __init__(self, config=None, core=None):
        self.process = CopasiProcess(config=self.config.get('process_config'))
        self.params_to_scan: List = self.config.get('parameters', [])
        self.n_iterations = self.config['n_iterations']
        self.iter_start = self.config.get('iter_start', 0.0)
        self.iter_stop = self.config['iter_stop']


    def initial_state(self):
        return {
            str(n): self.process.initial_state()
            for n in range(self.n_iterations)}

    def inputs(self):
        return self.process.inputs()

    def outputs(self):
        return {
            str(n): self.process.outputs()
            for n in range(self.n_iterations)}

    def update(self, input):
        """Here is where the method of perturbation differs: deterministic will use
            `np.linspace(...`
        """
        # set up parameters
        results = {}
        scan_range = np.linspace(
            start=self.iter_start,
            stop=self.iter_stop,
            num=self.n_iterations).tolist()

        for index, perturbed in enumerate(scan_range):
            interval = input['time']
            for param in self.params_to_scan:
                if 'global' in param.scope:
                    input_key = 'model_parameters'
                elif 'species' in param.scope:
                    input_key = 'floating_species'
                else:
                    raise ValueError('Only global or species -level parameter scanning is currently supported.')

                for input_id, input_value in input[input_key]:
                    if param.name in input_id:
                        input[input_key][input_id] = perturbed

                r = self.process.update(
                    inputs=input,
                    interval=interval)
                results[str(index)] = r
        return results


class StochasticParameterScan(ParameterScan):
    """Analogous to Monte Carlo perturbations"""
    config_schema = {
        'n_iterations': 'int',
        'perturbation_magnitude': 'float',
        'parameters': 'tree[string]',
        'process_config': 'tree[string]'}

    pass



import os
from process_bigraph import Composite
from biosimulators_processes import CORE
from tests import ProcessUnitTest


def test_process():
    # 1. define the instance of the Composite(in this case singular) by its schema
    sbml_model_files_root = 'biosimulators_processes/model_files/sbml'
    model_name = 'vilar-discrete-SSA.xml'  # 'BIOMD0000000061_url.xml'
    model_fp = os.path.join(sbml_model_files_root, model_name)

    composite_document = {
        'tellurium': {
            '_type': 'step',
            'address': 'local:get_sbml',
            'config': {
                'biomodel_id': model_fp,
            },
            'inputs': {
                'time': ['start_time_store'],
                'run_time': ['run_time_store'],
                'floating_species': ['floating_species_store'],
                'boundary_species': ['boundary_species_store'],
                'model_parameters': ['model_parameters_store'],
                'reactions': ['reactions_store'],
                # 'interval': ['interval_store'],
            },
            'outputs': {
                'results': ['results_store'],
            }
        },
        'emitter': {
            '_type': 'step',
            'address': 'local:ram-emitter',
            'config': {
                'emit': {
                    'floating_species': 'tree[float]',
                    'time': 'float',
                },
            },
            'inputs': {
                'floating_species': ['floating_species_store'],
                'time': ['start_time_store'],
            }
        }
    }

    # 2. make the composite
    workflow = Composite(
        config={'state': composite_document},
        core=CORE)

    # 3. run
    update = workflow.run(10)

    # 4. gather results
    results = workflow.gather_results()
    print(f'RESULTS: {pf(results)}')

    return ProcessUnitTest(instance_doc=instance)


if __name__ == '__main__':
    test_process()



# setup utils
from typing import Callable

import numpy as np
import rustworkx as rx
from qiskit.circuit.library import QAOAAnsatz


def rustworkx_graph_setup(w: list[list[float]]):
    num_nodes = len(w)
    w = np.array(w)

    # Create graph from adjacency matrix
    edges = []
    for i in range(num_nodes):
        for j in range(num_nodes):
            if w[i, j] > 0:  # Add edges where weight > 0
                edges.append((i, j, w[i, j]))

    G = rx.PyGraph()
    G.add_nodes_from(range(num_nodes))  # Add nodes
    G.add_edges_from(edges)  # Add edges with weights
    return G


def configure_runtime(circuit: QAOAAnsatz):
    import os
    from qiskit_ibm_runtime import QiskitRuntimeService
    from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager

    QiskitRuntimeService.save_account(channel="ibm_quantum", token=os.getenv("IBM_QUANTUM_TOKEN"), overwrite=True, set_as_default=True)
    service = QiskitRuntimeService(channel='ibm_quantum')
    backend = service.least_busy(min_num_qubits=127)
    print(backend)

    # Create pass manager for transpilation
    pm = generate_preset_pass_manager(optimization_level=3,
                                      backend=backend)

    candidate_circuit = pm.run(circuit)
    candidate_circuit.draw('mpl', fold=False, idle_wires=False)

    return backend, pm, candidate_circuit


def get_node_pairs(nodes: list) -> list[tuple]:
    from itertools import combinations
    return list(combinations(nodes, 2))


def random_weight() -> float:
    n = np.random.rand()
    return n ** n / n if n > 0.5 else n ** n * n


def get_weights(node_pairs: list[tuple], generator: Callable, as_tuple: bool = False) -> list[tuple[int, int, float]]:
    # node_list = [(0, 1), (0, 2), (0, 4), (1, 2), (2, 3), (3, 4)]
    return list(map(
        lambda nodes: (*nodes, generator()) if as_tuple else [*nodes, generator()],
        node_pairs
    ))


def get_random_weights(node_pairs: list[tuple]):
    return get_weights(node_pairs, random_weight)


def random_connection():
    n = np.random.rand()
    return 1.0 if n > 0.5 else 0.0


def random_adjacency_matrix(num_nodes: int):
    r = range(num_nodes)
    w = [[random_connection() for _ in r] for _ in r]
    return w


def draw_graph(graph, node_size=600):
    from rustworkx.visualization import mpl_draw as _draw_graph
    return _draw_graph(graph, node_size=node_size, with_labels=True)


def initialize_graph_k(n_nodes_k: int) -> rx.PyGraph:
    """
    Initialize new graph for iteration k.

    :params n_nodes_k: (`int`) Number of nodes (variables) at iteration k for the N-var qaoa solution.

    :rtype: `rustworkx.PyGraph` instance parameterized by n nodes.
    """
    graph_k = rx.PyGraph()

    # add num_nodes for iteration k (n_variables for k)
    nodes_k = np.arange(0, n_nodes_k, 1).tolist()
    graph_k.add_nodes_from(nodes_k)

    # add edge list for iteration k
    node_pairs = get_node_pairs(nodes_k)
    edges_k = get_weights(node_pairs, random_weight)
    graph_k.add_edges_from(edges_k)
    return graph_k



"""
Membrane process using forward euler gradient descent integration.
"""

import inspect
import os
import shutil
from functools import partial
from pathlib import Path
from typing import Dict, Union, List, Tuple
import tempfile as tmp

import numpy as np
import pymem3dg as dg
import pymem3dg.boilerplate as dgb
from netCDF4 import Dataset
from process_bigraph import Process, ProcessTypes, Composite, pp

from bsp.utils.base_utils import new_document
from bsp.utils.membrane_utils import extract_data, parse_ply


class MembraneProcess(Process):
    config_schema = {
        'mesh_file': 'string',
        'geometry': {
            'type': 'string',  # if used, ie; 'icosphere'
            'params': 'tree'  # params required for aforementioned shape type
        },
        'tension_model': {
            'modulus': 'float',
            'preferredArea': 'float'
        },
        'osmotic_model': {
            'preferredVolume': 'float',
            'reservoirVolume': 'float',
            'strength': 'float'  # what units is this in??
        },
        'parameters': 'tree',
        'save_period': 'integer',
        'tolerance': 'float',
        'characteristic_time_step': 'integer',
        'total_time': 'integer',
        'growth_coefficient': 'integer',  # growth to volume coeff ?
    }

    def __init__(self, config: Dict[str, Union[Dict[str, float], str]] = None, core: ProcessTypes = None):
        super().__init__(config, core)
        # get simulation params
        self.save_period = self.config.get("save_period", 100)  # interval at which sim state is saved to disk/recorded
        self.tolerance = self.config.get("tolerance", 1e-11)
        self.characteristic_time_step = self.config.get("characteristic_time_step", 2)
        self.total_time = self.config.get("total_time", 100)
        # that should be given to the composite?:
        self.total_time = self.config.get("total_time", 1000)
        self.default_growth_coefficient = self.config.get("growth_coefficient", 100)

        # parse input of either mesh file or geometry spec
        mesh_file = self.config.get("mesh_file")
        geometry = self.config.get("geometry")
        if mesh_file:
            self.initial_faces, self.initial_vertices = parse_ply(mesh_file)
        else:
            shape = self.config['geometry']['type']
            mesh_constructor = getattr(dg, f'get{shape.replace(shape[0], shape[0].upper())}')
            self.initial_faces, self.initial_vertices = mesh_constructor(**self.config['geometry']['params'])

        # set parameters TODO: should this be dynamic?
        self.parameters = dg.Parameters()
        init_param_spec = self.config.get("parameters")
        if init_param_spec:
            for attribute_name, attribute_spec in init_param_spec.items():  # ie: adsorption, aggregation, bending, etc
                attribute = getattr(self.parameters, attribute_name)
                for name, value in attribute_spec.items():
                    setattr(attribute, name, value)

        self.default_preferred_volume = self.config["osmotic_model"]["preferredVolume"],  # make input port here if value has changed (fba)
        self.default_reservoir_volume = self.config["osmotic_model"]["reservoirVolume"],  # output port
        self.default_osmotic_strength = self.config["osmotic_model"]["strength"]

        self.solver_factory = partial(
            dg.Euler,
            characteristicTimeStep=self.characteristic_time_step,
            totalTime=self.total_time,  # interval?
            tolerance=self.tolerance
            # system=system_k,
            # savePeriod=interval,
            # outputDirectory=str(output_dir)
        )

    def initial_state(self):
        initial_geometry = {
            "faces": self.initial_faces.tolist(),
            "vertices": self.initial_vertices.tolist(),
        }
        initial_velocities = [[0.0] for _ in self.initial_vertices]
        return {
            "geometry": initial_geometry,
            "velocities": initial_velocities,
        }

    def inputs(self):
        return {
            'geometry': 'GeometryType',  # faces: {[[]]}, vertices: [[]]}
            'species_concentrations': 'tree[float]',  # like {'species_concentrations': {'osmotic': {}, etc}}
            'fluxes': 'tree[float]',
            'growth_coefficient': 'float',
            # 'preferred_volume': 'float',  # param for osmotic pressure model
            # 'current_volume': 'float',  # param for osmotic pressure model
            # 'tension_model': 'TensionModelParams',  # todo: can we parse this from the previous two fields?
            # 'species_concentrations': 'tree[float]',  # particularly, preferred volume
            # 'fluxes': 'tree[float]',  # parsing preferred volume from dfba outputs?
        }

    def outputs(self):
        return {
            'geometry': 'GeometryType',
            'velocities': 'VelocitiesType',
        }

    def update(self, state, interval):
        # parse kth-1 geometry output
        previous_geometry = state.get("geometry")
        input_faces = previous_geometry["faces"]
        input_vertices = previous_geometry["vertices"]
        previous_faces = np.array(input_faces) if isinstance(input_faces, list) else input_faces
        previous_vertices = np.array(input_vertices) if isinstance(input_vertices, list) else input_vertices

        # instantiate new geometry for k
        geometry_k = dg.Geometry(previous_faces, previous_vertices)

        # calculate current volume from kth inputs
        input_fluxes = state.get("fluxes")
        input_osmotic_species_concentrations = state.get("species_concentrations").get('osmotic')
        total_osmotic_concentration = sum(input_osmotic_species_concentrations.values())
        n_solutes = sum(input_fluxes[reaction] for reaction in input_fluxes if "transport" in reaction)
        volume_k = n_solutes / total_osmotic_concentration

        # calculate preferred volume
        # nutrient_concentration = 0.5
        # Km = 0.3  # Michaelis constant
        # k_base = 1000  # Baseline coefficient
        # growth_to_volume_coeff = k_base * (1 + nutrient_concentration / (Km + nutrient_concentration))
        input_growth_coeff = state.get("growth_coefficient", self.default_growth_coefficient)
        growth_flux = input_fluxes.get("biomass_reaction", 0)
        growth_volume = input_growth_coeff * growth_flux
        osmotic_volume = volume_k  # TODO: this assumes the osmotic contribution directly determines the volume of the cell at equilibrium
        preferred_volume_k = volume_k + osmotic_volume * 1e12 + growth_volume  # converts to L TODO: how should we handle units?
        osmotic_strength_k = state.get("osmotic_strength", self.default_osmotic_strength)

        # set osmotic model callback
        osmotic_model_k = partial(
            dgb.preferredVolumeOsmoticPressureModel,
            preferredVolume=preferred_volume_k,  # make input port here if value has changed (fba)
            volume=volume_k,  # output port
            strength=osmotic_strength_k,
        )

        # calculate area and preferred area from volume params
        radius_k = (3 * volume_k / (4 * np.pi)) ** (1/3)
        area_k = 4 * np.pi * radius_k**2
        preferred_radius_k = (3 * preferred_volume_k / (4 * np.pi)) ** (1 / 3)
        preferred_area_k = 4 * np.pi * preferred_radius_k ** 2

        # set tension model callback
        tension_model_k = partial(
            dgb.preferredAreaSurfaceTensionModel,
            area=area_k,
            modulus=self.config["tension_model"]["modulus"],
            preferredArea=preferred_area_k,
        )

        # instantiate the params with the class pressure models
        param_config = state.get("parameters", self.config["parameters"])
        parameters_k = new_parameters(param_spec=param_config)
        parameters_k.tension.form = tension_model_k
        parameters_k.osmotic.form = osmotic_model_k

        # set up system from parsed geometry and parameters for kth
        system_k = dg.System(
            geometry=geometry_k,
            parameters=parameters_k
        )
        system_k.initialize()

        # mk temp dir to save and parse kth output io
        output_dir = Path(tmp.mkdtemp())

        # set up solver
        # time_points = np.arange(start=interval - 1, stop=(total_time + characteristic_step), step=characteristic_step
        # system=system_k,
        # savePeriod=interval,
        # outputDirectory=str(output_dir)
        # fe = dg.Euler(
        #     system=system_k,
        #     characteristicTimeStep=self.characteristic_time_step,
        #     savePeriod=interval,
        #     totalTime=1,  # interval?
        #     tolerance=self.tolerance,
        #     outputDirectory=str(output_dir)
        # )
        # fe.ifPrintToConsole = True
        # fe.ifOutputTrajFile = True

        fe_k = self.solver_factory(
            system=system_k,
            savePeriod=interval,
            outputDirectory=str(output_dir)
        )
        fe_k.ifPrintToConsole = True
        fe_k.ifOutputTrajFile = True

        # run solver and extract data
        success = fe_k.integrate()  # or should this be fe.step(interval)?
        output_path = str(output_dir / "traj.nc")
        data = Dataset(output_path, 'r')

        # get velocities
        velocities_k = extract_data(dataset=data, data_name="velocities", return_last=False)

        # get faces (do we need this?)
        faces_k = extract_data(dataset=data, data_name="topology", return_last=False)

        # get vertices
        vertices_k = extract_data(dataset=data, data_name="coordinates", return_last=False)

        # parse parameters for iteration
        # param_data_k = parse_parameters(parameters=parameters_k)

        # clean up temporary files
        shutil.rmtree(str(output_dir))

        geometry_out = {
            "vertices": vertices_k,
            "faces": faces_k,
        }

        return {
            "geometry": geometry_out,
            # "parameters": param_data_k,
            "velocities": velocities_k
        }


def new_parameters(param_spec: Dict) -> dg.Parameters:
    parameters = dg.Parameters()
    if param_spec:
        for attribute_name, attribute_spec in param_spec.items():  # ie: adsorption, aggregation, bending, etc
            attribute = getattr(parameters, attribute_name)
            for name, value in attribute_spec.items():
                setattr(attribute, name, value)

    return parameters


def test_membrane_process():
    # get vertices this will be divisible by 3 and thus is a time indexed flat list of tuples(xyz) for each vertex in the mesh
    from bsp import app_registrar

    # define config and port vals
    test_config = {
        'mesh_file': '/Users/alexanderpatrie/Desktop/repos/biosimulator-processes/tests/fixtures/sample_meshes/oblate.ply',
        'tension_model': {
            'modulus': 0.1,
            'preferredArea': 12.4866
        },
        'osmotic_model': {
            'preferredVolume': 0.7 * np.pi * 4 / 3,
            'reservoirVolume': 0,
            'strength': 0.02
        },
        'parameters': {
            'bending': {
                'Kbc': 8.22e-5
            }
        },
        'save_period': 100,
        'tolerance': 1e-11,
        'characteristic_time_step': 2
    }

    port_schema = {
        "geometry": {
            "faces": ["faces_store"],
            "vertices": ["vertices_store"],
        },
        "parameters": ["parameters_store"],
        "velocities": ["velocities_store"]
    }

    # register process
    app_registrar.register_process("membrane-process", MembraneProcess, verbose=True)

    # create spec
    document = new_document(
        name='membrane',
        address='membrane-process',
        _type='process',
        config=test_config,
        inputs=port_schema,
        outputs=port_schema,
        add_emitter=True
    )

    # configure and run composite
    sim = Composite(
        config={'state': document},
        core=app_registrar.core
    )
    total_time = 10000
    sim.run(total_time)

    # get results
    data = sim.gather_results()
    pp(data)





from typing import Tuple, Callable, Union, List, Dict
from dataclasses import dataclass
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from process_bigraph.composite import Step, Composite, RAMEmitter


def parse_composition_results(composition: Composite) -> Dict[float, Dict[str, Union[Dict, int, float]]]:
    """Return a dictionary in which the outer keys are each member of the timescale, which is parsed
        from the `process_bigraph.Composite.state['global_time']`. Each outer key's children represent the
        data returned from the emitter for ports at each time step.

        Args:
            composition:`process_bigraph.Composite`: Composite instance from which data is being retrieved.

        Returns:
            `Dict[float, Dict[str, Union[Dict, int, float]]]`
            A dictionary of keys: result_values.
    """
    results = composition.gather_results()
    result_vals = list(results.values())[0]
    timescale = [
        float(n)
        for n in range(int(composition.state['global_time']) + 1) # python ranges are not as they appear ;)
    ]
    assert len(timescale) == len(result_vals)  # src check: is there a better way to perform the check?
    return dict(zip(timescale, result_vals))


@dataclass
class PlotLabelConfig:
    title: str
    x_label: str
    y_label: str


class ResultsAnimation:
    def __init__(self,
                 x_data: np.array,
                 y_data: np.array,
                 plot_type_func: Callable,
                 plot_label_config: PlotLabelConfig,
                 **config):
        self.x_data = x_data
        self.y_data = y_data
        self.plot_type_func = plot_type_func
        self.plot_label_config = plot_label_config
        self.config = config

    def _set_axis_limits(self, ax_limit_func, data):
        return ax_limit_func(data.min(), data.max())

    def _prepare_animation(self, t):
        plt.cla()
        self.plot_type_func(
            self.x_data,
            self.y_data,
            color=self.config.get('color', 'blue')
        )

        self._set_axis_limits(plt.xlim, self.x_data)
        self._set_axis_limits(plt.ylim, self.y_data)
        plt.axhline(0, color='grey', lw=0.5)
        plt.axvline(0, color='grey', lw=0.5)
        plt.title(self.plot_label_config.title)
        plt.xlabel(self.plot_label_config.x_label)
        plt.ylabel(self.plot_label_config.y_label)

    def _create_animation_components(self) -> Tuple:
        """
        Creates a matplotlib subplot setup for animation.

        Returns:
            Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]: A tuple containing the figure and axes objects.
        """
        plt.rcParams["animation.html"] = "jshtml"
        plt.rcParams['figure.dpi'] = 150  # TODO: add this to config
        plt.ioff()
        fig, ax = plt.subplots()
        return fig, ax

    def run(self, num_frames: int):
        """
        Creates and runs the animation.

        Args:
            num_frames (int): Number of frames to animate

        Returns:
            matplotlib.animation.FuncAnimation: The animation object.
        """
        fig, ax = self._create_animation_components()
        return FuncAnimation(fig, self._prepare_animation, frames=num_frames)



class CompositionPlotter(Step):

    data: Dict[float, Dict[str, Union[Dict, int, float]]]
    emitter: RAMEmitter
    config_schema = {
        'emitter': 'step',
        'plot_counts': {
            '_type': 'boolean',
            '_default': False
        }
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

        self.emitter = self.config.get('emitter')
        assert self.emitter is not None, 'You must pass a ram-emitter instance into this plotter'
        self.data = self.emitter.history
        self.species_names = []
        for timestamp, result in self.data.items():
            names = list(result['floating_species_concentrations'].keys())
            for n in names:
                self.species_names.append(n)

        self.species_names = list(set(self.species_names))

    def _parse_data(self, species_names: List):
        fig, axes = plt.subplots(nrows=len(species_names), ncols=2, figsize=(15, 5 * len(species_names)))
        fig.suptitle('Species Counts and Concentrations Over Time')

        # Process each species
        for index, species in enumerate(species_names):
            times = list(data.keys())
            counts = [data[time]['floating_species_counts'][species] for time in times]
            concentrations = [data[time]['floating_species_concentrations'][species] for time in times]

            # Plot counts
            ax1 = axes[index, 0]
            ax1.plot(times, counts, label=f'Counts of {species}')
            ax1.set_title(f'Counts of {species}')
            ax1.set_xlabel('Time')
            ax1.set_ylabel('Counts')
            ax1.legend()

            # Plot concentrations
            ax2 = axes[index, 1]
            ax2.plot(times, concentrations, label=f'Concentration of {species}')
            ax2.set_title(f'Concentration of {species}')
            ax2.set_xlabel('Time')
            ax2.set_ylabel('Concentration')
            ax2.legend()

        plt.tight_layout(rect=[0, 0, 1, 0.96])
        plt.show()

    def inputs(self):
        return {'emitter': 'step'}

    def outputs(self):
        return {}

    def update(self, state):
        return self._parse_data(self.species_names)




class Plotter2d(Step):
    config_schema = {
        'duration': 'any',  # TODO: change this
        'process': 'string',
        'species_context': {
            '_type': 'string',
            '_default': 'concentrations'
        }
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)
        context = self.config['species_context']
        self.context_key = f'floating_species_{context}'
        self.species_store = [f'{self.context_key}_store']


    def inputs(self):
        return {
            self.context_key: 'tree[float]'
        }

    def outputs(self):
        return {
            self.context_key: 'tree[float]'
        }

    def update(self, state):
        """print(f'GOt the state: {state}')
        results = state[self.context_key]
        species_names = list(results.keys())
        timescale = list(range(self.config['duration'] + 1))
        for name in species_names:
            y_data = results[name]
            self.plot_output(
                x_data=timescale,
                y_data=y_data,
                species=name,
                x_label='Time',
                y_label=self.context_key
            )
        return {}"""
        print('----', state)
        print('----', self.config['process'])
        return {}


    @classmethod
    def plot_output(
            cls,
            x_data: Union[List, np.ndarray],
            y_data: Union[List, np.ndarray],
            **label_config
    ) -> None:
        """Plot arbitrary output.

            Args:
                x_data: array or list of data for x axis.
                y_data: array or list of data for y axis.
                **label_config: kwargs include--> title, x_label, y_label
        """
        plt.figure(figsize=(8, 5))
        plt.plot(x_data, y_data, marker='o', linestyle='-', color='b', label=label_config.get('species'))
        plt.title(label_config.get('title'))
        plt.xlabel(label_config.get('x_label'))
        plt.ylabel(label_config.get('y_label'))
        plt.legend()
        plt.grid(True)
        plt.show()

    @classmethod
    def plot_single_output(
            cls,
            timescale: Union[List, np.array],
            data: Union[List, np.array],
            species_name: str,
            plot_concentration=True
    ) -> None:
        """Plotting function to plot the output of a SINGLE species' concentrations over a timescale for the output
            of a deterministic time course simulation.

            Args:
                timescale:`Union[List, np.array]`: list containing each time step.
                data:`Union[List, np.array]`: output data mapped to each timescale.
                species_name:`str`: Name of the species that you are plotting.
                plot_concentration:`bool`: Whether you are plotting concentrations for data. Effects
                    plot labels. Defaults to `True`.
        """
        plt.figure(figsize=(8, 5))
        plt.plot(timescale, data, marker='o', linestyle='-', color='b', label=species_name)
        plt.title(f'{species_name} over time')
        plt.xlabel('Time')
        plt.ylabel('Concentration' if plot_concentration else 'Counts')
        plt.legend()
        plt.grid(True)
        plt.show()


